import tensorflow as tf

# constant conversion matrices between color spaces: https://gist.github.com/bikz05/6fd21c812ef6ebac66e1
_rgb2xyz = tf.constant([[0.412453, 0.357580, 0.180423],
                        [0.212671, 0.715160, 0.072169],
                        [0.019334, 0.119193, 0.950227]])
                        
_white = tf.constant([0.95047, 1., 1.08883])

def rgb2lab(rgb):
    arr = tf.cast(rgb, tf.float32)

    # convert rgb -> xyz color domain
    mask = arr > 0.04045
    not_mask = tf.math.logical_not(mask)
    arr = tf.tensor_scatter_nd_update(arr, tf.where(mask), tf.math.pow((tf.boolean_mask(arr, mask) + 0.055) / 1.055, 2.4))
    arr = tf.tensor_scatter_nd_update(arr, tf.where(not_mask), tf.boolean_mask(arr, not_mask) / 12.92)

    xyz = arr @ tf.cast(tf.transpose(_rgb2xyz), arr.dtype)

    # scale by CIE XYZ tristimulus values of the reference white point
    arr = tf.identity(xyz)
    arr = arr / tf.cast(_white, xyz.dtype)

    # nonlinear distortion and linear transformation
    mask = arr > 0.008856
    not_mask = tf.math.logical_not(mask)
    arr = tf.tensor_scatter_nd_update(arr, tf.where(mask), tf.math.pow(tf.boolean_mask(arr, mask), 1.0 / 3.0))
    arr = tf.tensor_scatter_nd_update(arr, tf.where(not_mask), 7.787 * tf.boolean_mask(arr, not_mask) + 16 / 116)

    # get each channel as individual tensors
    x, y, z = arr[..., 0], arr[..., 1], arr[..., 2]

    # vector scaling
    L = (116. * y) - 16.
    a = 500.0 * (x - y)
    b = 200.0 * (y - z)

    # OpenCV format
    L *= 2.55
    a += 128
    b += 128

    # finally, get LAB color domain
    return tf.stack([L, a, b], axis=-1)



import numpy as np
from torchstain.base.normalizers import HENormalizer
from torchstain.numpy.utils.rgb2lab import rgb2lab
from torchstain.numpy.utils.lab2rgb import lab2rgb
from torchstain.numpy.utils.split import csplit, cmerge, lab_split, lab_merge
from torchstain.numpy.utils.stats import get_mean_std, standardize

"""
Source code adapted from:
https://github.com/DigitalSlideArchive/HistomicsTK/blob/master/histomicstk/preprocessing/color_normalization/reinhard.py
https://github.com/Peter554/StainTools/blob/master/staintools/reinhard_color_normalizer.py
"""
class NumpyReinhardNormalizer(HENormalizer):
    def __init__(self, method=None):
        super().__init__()
        self.method = method
        self.target_mus = None
        self.target_stds = None
    
    def fit(self, target):
        # normalize
        target = target.astype("float32") / 255

        # convert to LAB
        lab = rgb2lab(target)

        # get summary statistics
        stack_ = np.array([get_mean_std(x) for x in lab_split(lab)])
        
        self.target_means = stack_[:, 0]
        self.target_stds = stack_[:, 1]

    def normalize(self, I):
        # normalize
        I = I.astype("float32") / 255
        
        # convert to LAB
        lab = rgb2lab(I)
        labs = lab_split(lab)

        # get summary statistics from LAB
        stack_ = np.array([get_mean_std(x) for x in labs])
        
        mus = stack_[:, 0]
        stds = stack_[:, 1]

        # normalize
        if self.method is None:
            # standardize intensities channel-wise and normalize using target mus and stds
            result = [standardize(x, mu_, std_) * std_T + mu_T for x, mu_, std_, mu_T, std_T \
                in zip(labs, mus, stds, self.target_means, self.target_stds)]

        elif self.method == "modified":
            # calculate q
            q = (self.target_stds[0] - stds[0]) / self.target_stds[0]
            q = 0.05 if q <= 0 else q

            # normalize each channel independently
            l_norm = mus[0] + (labs[0] - mus[0]) * (1 + q)
            a_norm = self.target_means[1] + (labs[1] - mus[1])
            b_norm = self.target_means[2] + (labs[2] - mus[2])

            result = [l_norm, a_norm, b_norm]

        else:
            raise ValueError("Unsupported 'method' was chosen. Choose either {None, 'modified'}.")

        # rebuild LAB
        lab = lab_merge(*result)

        # convert back to RGB from LAB
        lab = lab2rgb(lab)

        # rescale to [0, 255] uint8
        return (lab * 255).astype("uint8")



import numpy as np
from torchstain.base.augmentors import HEAugmentor

"""
Source code adapted from: https://github.com/schaugf/HEnorm_python
Original implementation: https://github.com/mitkovetta/staining-normalization
"""
class NumpyMacenkoAugmentor(HEAugmentor):
    def __init__(self, sigma1=0.2, sigma2=0.2):
        super().__init__()

        self.sigma1 = sigma1
        self.sigma2 = sigma2

        self.I = None

        self.HERef = np.array([[0.5626, 0.2159],
                               [0.7201, 0.8012],
                               [0.4062, 0.5581]])
        self.maxCRef = np.array([1.9705, 1.0308])

    def __convert_rgb2od(self, I, Io=240, beta=0.15):
        # calculate optical density
        OD = -np.log((I.astype(float) + 1) / Io)

        # remove transparent pixels
        ODhat = OD[~np.any(OD < beta, axis=1)]

        return OD, ODhat

    def __find_HE(self, ODhat, eigvecs, alpha):
        #project on the plane spanned by the eigenvectors corresponding to the two
        # largest eigenvalues
        That = ODhat.dot(eigvecs[:,1:3])

        phi = np.arctan2(That[:,1],That[:,0])

        minPhi = np.percentile(phi, alpha)
        maxPhi = np.percentile(phi, 100-alpha)

        vMin = eigvecs[:, 1:3].dot(np.array([(np.cos(minPhi), np.sin(minPhi))]).T)
        vMax = eigvecs[:, 1:3].dot(np.array([(np.cos(maxPhi), np.sin(maxPhi))]).T)

        # a heuristic to make the vector corresponding to hematoxylin first and the
        # one corresponding to eosin second
        if vMin[0] > vMax[0]:
            HE = np.array((vMin[:,0], vMax[:,0])).T
        else:
            HE = np.array((vMax[:,0], vMin[:,0])).T

        return HE

    def __find_concentration(self, OD, HE):
        # rows correspond to channels (RGB), columns to OD values
        Y = np.reshape(OD, (-1, 3)).T

        # determine concentrations of the individual stains
        C = np.linalg.lstsq(HE, Y, rcond=None)[0]

        return C

    def __compute_matrices(self, I, Io, alpha, beta):
        I = I.reshape((-1, 3))

        OD, ODhat = self.__convert_rgb2od(I, Io=Io, beta=beta)

        # compute eigenvectors
        _, eigvecs = np.linalg.eigh(np.cov(ODhat.T))

        HE = self.__find_HE(ODhat, eigvecs, alpha)

        C = self.__find_concentration(OD, HE)

        # normalize stain concentrations
        maxC = np.array([np.percentile(C[0,:], 99), np.percentile(C[1,:], 99)])

        return HE, C, maxC

    def fit(self, I, Io=240, alpha=1, beta=0.15):
        HE, C, maxC = self.__compute_matrices(I, Io, alpha, beta)

        # keep these as we will use them for augmentation
        self.I = I
        self.HERef = HE
        self.CRef = C
        self.maxCRef = maxC
    
    def augment(self, Io=240, alpha=1, beta=0.15):
        I = self.I
        h, w, c = I.shape
        I = I.reshape((-1, 3))

        HE, C, maxC = self.__compute_matrices(I, Io, alpha, beta)

        maxC = np.divide(maxC, self.maxCRef)
        C2 = np.divide(C, maxC[:, np.newaxis])

        # introduce noise to the concentrations
        for i in range(C2.shape[0]):
            C2[i, :] *= np.random.uniform(1 - self.sigma1, 1 + self.sigma1)  # multiplicative
            C2[i, :] += np.random.uniform(-self.sigma2, self.sigma2)  # additative

        # recreate the image using reference mixing matrix
        Iaug = np.multiply(Io, np.exp(-self.HERef.dot(C2)))
        Iaug[Iaug > 255] = 255
        Iaug = np.reshape(Iaug.T, (h, w, c)).astype(np.uint8)

        return Iaug



import torch
from torchstain.torch.utils.rgb2lab import _rgb2xyz, _white

_xyz2rgb = torch.linalg.inv(_rgb2xyz)

def lab2rgb(lab):
    lab = lab.type(torch.float32)
    
    # rescale back from OpenCV format and extract LAB channel
    L, a, b = lab[0] / 2.55, lab[1] - 128, lab[2] - 128

    # vector scaling to produce X, Y, Z
    y = (L + 16.) / 116.
    x = (a / 500.) + y
    z = y - (b / 200.)

    # merge back to get reconstructed XYZ color image
    out = torch.stack([x, y, z], axis=0)

    # apply boolean transforms
    mask = out > 0.2068966
    not_mask = torch.logical_not(mask)
    out.masked_scatter_(mask, torch.pow(torch.masked_select(out, mask), 3))
    out.masked_scatter_(not_mask, (torch.masked_select(out, not_mask) - 16 / 116) / 7.787)

    # rescale to the reference white (illuminant)
    out = torch.mul(out, _white.type(out.dtype).unsqueeze(dim=-1).unsqueeze(dim=-1))

    # convert XYZ -> RGB color domain
    arr = torch.tensordot(out, torch.t(_xyz2rgb).type(out.dtype), dims=([0], [0]))
    mask = arr > 0.0031308
    not_mask = torch.logical_not(mask)
    arr.masked_scatter_(mask, 1.055 * torch.pow(torch.masked_select(arr, mask), 1 / 2.4) - 0.055)
    arr.masked_scatter_(not_mask, torch.masked_select(arr, not_mask) * 12.92)
    return torch.clamp(arr, 0, 1)



import tensorflow as tf
from torchstain.base.normalizers.he_normalizer import HENormalizer
from torchstain.tf.utils import cov, percentile, solveLS
import numpy as np
import tensorflow.keras.backend as K

"""
Source code ported from: https://github.com/schaugf/HEnorm_python
Original implementation: https://github.com/mitkovetta/staining-normalization
"""
class TensorFlowMacenkoNormalizer(HENormalizer):
    def __init__(self):
        super().__init__()

        self.HERef = tf.constant([[0.5626, 0.2159],
                                  [0.7201, 0.8012],
                                  [0.4062, 0.5581]])
        self.maxCRef = tf.constant([1.9705, 1.0308])

    def __convert_rgb2od(self, I, Io, beta):
        I = tf.transpose(I, [1, 2, 0])

        # calculate optical density
        OD = -tf.math.log((tf.cast(tf.reshape(I, [tf.math.reduce_prod(I.shape[:-1]), I.shape[-1]]), tf.float32) + 1) / Io)

        # remove transparent pixels
        ODhat = OD[~tf.math.reduce_any(OD < beta, axis=1)]

        return OD, ODhat

    def __find_HE(self, ODhat, eigvecs, alpha):
        # project on the plane spanned by the eigenvectors corresponding to the two
        # largest eigenvalues
        That = tf.linalg.matmul(ODhat, eigvecs)
        phi = tf.math.atan2(That[:, 1], That[:, 0])

        minPhi = percentile(phi, alpha)
        maxPhi = percentile(phi, 100 - alpha)

        vMin = tf.matmul(eigvecs, tf.expand_dims(tf.stack((tf.math.cos(minPhi), tf.math.sin(minPhi))), axis=-1))
        vMax = tf.matmul(eigvecs, tf.expand_dims(tf.stack((tf.math.cos(maxPhi), tf.math.sin(maxPhi))), axis=-1))

        # a heuristic to make the vector corresponding to hematoxylin first and the
        # one corresponding to eosin second
        HE = tf.where(vMin[0] > vMax[0], tf.concat((vMin, vMax), axis=1), tf.concat((vMax, vMin), axis=1))

        return HE

    def __find_concentration(self, OD, HE):
        # rows correspond to channels (RGB), columns to OD values
        Y = tf.transpose(OD)

        # determine concentrations of the individual stains
        return solveLS(HE, Y)[:2]

    def __compute_matrices(self, I, Io, alpha, beta):
        OD, ODhat = self.__convert_rgb2od(I, Io=Io, beta=beta)

        # compute eigenvectors
        _, eigvecs = tf.linalg.eigh(cov(tf.transpose(ODhat)))
        eigvecs = eigvecs[:, 1:3]

        HE = self.__find_HE(ODhat, eigvecs, alpha)

        C = self.__find_concentration(OD, HE)
        maxC = tf.stack([percentile(C[0, :], 99), percentile(C[1, :], 99)])

        return HE, C, maxC

    def fit(self, I, Io=240, alpha=1, beta=0.15):
        HE, _, maxC = self.__compute_matrices(I, Io, alpha, beta)

        self.HERef = HE
        self.maxCRef = maxC

    def normalize(self, I, Io=240, alpha=1, beta=0.15, stains=True):
        ''' Normalize staining appearence of H&E stained images

        Example use:
            see example.py

        Input:
            I: RGB input image: tensor of shape [C, H, W] and type uint8
            Io: (optional) transmitted light intensity
            alpha: percentile
            beta: transparency threshold
            stains: if true, return also H & E components

        Output:
            Inorm: normalized image
            H: hematoxylin image
            E: eosin image

        Reference:
            A method for normalizing histology slides for quantitative analysis. M.
            Macenko et al., ISBI 2009
        '''
        c, h, w = I.shape

        HE, C, maxC = self.__compute_matrices(I, Io, alpha, beta)

        # normalize stain concentrations
        C *= tf.expand_dims((self.maxCRef / maxC), axis=-1)

        # recreate the image using reference mixing matrix
        Inorm = Io * tf.math.exp(-tf.linalg.matmul(self.HERef, C))
        Inorm = tf.clip_by_value(Inorm, 0, 255)
        Inorm = tf.cast(tf.reshape(tf.transpose(Inorm), shape=(h, w, c)), tf.int32)

        H, E = None, None

        if stains:
            H = tf.math.multiply(Io, tf.math.exp(tf.linalg.matmul(-tf.expand_dims(self.HERef[:, 0], axis=-1), tf.expand_dims(C[0, :], axis=0))))
            H = tf.clip_by_value(H, 0, 255)
            H = tf.cast(tf.reshape(tf.transpose(H), shape=(h, w, c)), tf.int32)

            E = tf.math.multiply(Io, tf.math.exp(tf.linalg.matmul(-tf.expand_dims(self.HERef[:, 1], axis=-1), tf.expand_dims(C[1, :], axis=0))))
            E = tf.clip_by_value(E, 0, 255)
            E = tf.cast(tf.reshape(tf.transpose(E), shape=(h, w, c)), tf.int32)

        return Inorm, H, E



import tensorflow as tf

def csplit(I):
    return [I[..., i] for i in range(I.shape[-1])]

def cmerge(I1, I2, I3):
    return tf.stack([I1, I2, I3], axis=-1)

def lab_split(I):
    I = tf.cast(I, tf.float32)
    I1, I2, I3 = csplit(I)
    return I1 / 2.55, I2 - 128, I3 - 128

def lab_merge(I1, I2, I3):
    return cmerge(I1 * 2.55, I2 + 128, I3 + 128)



import tensorflow as tf
from torchstain.tf.utils.rgb2lab import _rgb2xyz, _white

_xyz2rgb = tf.linalg.inv(_rgb2xyz)

def lab2rgb(lab):
    lab = tf.cast(lab, tf.float32)
    
    # rescale back from OpenCV format and extract LAB channel
    L, a, b = lab[..., 0] / 2.55, lab[..., 1] - 128, lab[..., 2] - 128

    # vector scaling to produce X, Y, Z
    y = (L + 16.) / 116.
    x = (a / 500.) + y
    z = y - (b / 200.)

    # merge back to get reconstructed XYZ color image
    out = tf.stack([x, y, z], axis=-1)

    # apply boolean transforms
    mask = out > 0.2068966
    not_mask = tf.math.logical_not(mask)
    out = tf.tensor_scatter_nd_update(out, tf.where(mask), tf.pow(tf.boolean_mask(out, mask), 3))
    out = tf.tensor_scatter_nd_update(out, tf.where(not_mask), (tf.boolean_mask(out, not_mask) - 16 / 116) / 7.787)

    # rescale to the reference white (illuminant)
    out = out * tf.cast(_white, out.dtype)
    
    # convert XYZ -> RGB color domain
    arr = tf.identity(out)
    arr = arr @ tf.transpose(_xyz2rgb)
    mask = arr > 0.0031308
    not_mask = tf.math.logical_not(mask)
    arr = tf.tensor_scatter_nd_update(arr, tf.where(mask), 1.055 * tf.pow(tf.boolean_mask(arr, mask), 1 / 2.4) - 0.055)
    arr = tf.tensor_scatter_nd_update(arr, tf.where(not_mask), tf.boolean_mask(out, not_mask) * 12.92)
    return tf.clip_by_value(arr, 0, 1)



def MacenkoAugmentor(backend='torch', sigma1=0.2, sigma2=0.2):
    if backend == 'numpy':
        from torchstain.numpy.augmentors import NumpyMacenkoAugmentor
        return NumpyMacenkoAugmentor(sigma1=sigma1, sigma2=sigma2)
    elif backend == "torch":
        from torchstain.torch.augmentors import TorchMacenkoAugmentor
        return TorchMacenkoAugmentor(sigma1=sigma1, sigma2=sigma2)
    elif backend == "tensorflow":
        from torchstain.tf.augmentors import TensorFlowMacenkoAugmentor
        return TensorFlowMacenkoAugmentor(sigma1=sigma1, sigma2=sigma2)
    else:
        raise Exception(f'Unknown backend {backend}')



import tensorflow as tf

def solveLS(A, B):
    q_full, _ = tf.linalg.qr(A, full_matrices=True)
    ret1 = tf.linalg.lstsq(A, B)
    ret2 = tf.linalg.lstsq(q_full, B)
    return tf.concat([ret1, ret2[ret1.shape[0]:]], axis=0)



from torchstain.torch.normalizers.macenko import TorchMacenkoNormalizer
from torchstain.torch.normalizers.multitarget import MultiMacenkoNormalizer
from torchstain.torch.normalizers.reinhard import TorchReinhardNormalizer



def MacenkoNormalizer(backend='torch'):
    if backend == 'numpy':
        from torchstain.numpy.normalizers import NumpyMacenkoNormalizer
        return NumpyMacenkoNormalizer()
    elif backend == "torch":
        from torchstain.torch.normalizers.macenko import TorchMacenkoNormalizer
        return TorchMacenkoNormalizer()
    elif backend == "tensorflow":
        from torchstain.tf.normalizers.macenko import TensorFlowMacenkoNormalizer
        return TensorFlowMacenkoNormalizer()
    else:
        raise Exception(f'Unknown backend {backend}')



from torchstain.numpy.utils.rgb2lab import rgb2lab
from torchstain.numpy.utils.lab2rgb import lab2rgb
import numpy as np
import cv2
import os

def test_rgb_lab():
    size = 1024
    curr_file_path = os.path.dirname(os.path.realpath(__file__))
    img = cv2.resize(cv2.cvtColor(cv2.imread(os.path.join(curr_file_path, "../data/source.png")), cv2.COLOR_BGR2RGB), (size, size))
    
    # rgb2lab expects data to be float32 in range [0, 1]
    img = img / 255

    # convert from RGB to LAB and back again to RGB
    reconstructed_img = lab2rgb(rgb2lab(img))

    # assess if the reconstructed image is similar to the original image
    np.testing.assert_almost_equal(np.mean(np.abs(reconstructed_img - img)), 0.0, decimal=4, verbose=True)



import torch
from torchstain.torch.utils import cov, percentile
"""
Implementation of the multi-target normalizer from the paper: https://arxiv.org/pdf/2406.02077
"""
class MultiMacenkoNormalizer:
    def __init__(self, norm_mode='avg-post'):
        self.norm_mode = norm_mode
        self.HERef = torch.tensor([[0.5626, 0.2159],
                                   [0.7201, 0.8012],
                                   [0.4062, 0.5581]])
        self.maxCRef = torch.tensor([1.9705, 1.0308])
        self.updated_lstsq = hasattr(torch.linalg, 'lstsq')
        
    def __convert_rgb2od(self, I, Io, beta):
        I = I.permute(1, 2, 0)
        OD = -torch.log((I.reshape((-1, I.shape[-1])).float() + 1)/Io)
        ODhat = OD[~torch.any(OD < beta, dim=1)]
        return OD, ODhat

    def __find_phi_bounds(self, ODhat, eigvecs, alpha):
        That = torch.matmul(ODhat, eigvecs)
        phi = torch.atan2(That[:, 1], That[:, 0])

        minPhi = percentile(phi, alpha)
        maxPhi = percentile(phi, 100 - alpha)

        return minPhi, maxPhi

    def __find_HE_from_bounds(self, eigvecs, minPhi, maxPhi):
        vMin = torch.matmul(eigvecs, torch.stack((torch.cos(minPhi), torch.sin(minPhi)))).unsqueeze(1)
        vMax = torch.matmul(eigvecs, torch.stack((torch.cos(maxPhi), torch.sin(maxPhi)))).unsqueeze(1)

        HE = torch.where(vMin[0] > vMax[0], torch.cat((vMin, vMax), dim=1), torch.cat((vMax, vMin), dim=1))

        return HE

    def __find_HE(self, ODhat, eigvecs, alpha):
        minPhi, maxPhi = self.__find_phi_bounds(ODhat, eigvecs, alpha)
        return self.__find_HE_from_bounds(eigvecs, minPhi, maxPhi)

    def __find_concentration(self, OD, HE):
        Y = OD.T
        if not self.updated_lstsq:
            return torch.lstsq(Y, HE)[0][:2]
        return torch.linalg.lstsq(HE, Y)[0]

    def __compute_matrices_single(self, I, Io, alpha, beta):
        OD, ODhat = self.__convert_rgb2od(I, Io=Io, beta=beta)

        _, eigvecs = torch.symeig(cov(ODhat.T), eigenvectors=True)
        eigvecs = eigvecs[:, [1, 2]]

        HE = self.__find_HE(ODhat, eigvecs, alpha)

        C = self.__find_concentration(OD, HE)
        maxC = torch.stack([percentile(C[0, :], 99), percentile(C[1, :], 99)])

        return HE, C, maxC

    def fit(self, Is, Io=240, alpha=1, beta=0.15):
        if self.norm_mode == 'avg-post':
            HEs, _, maxCs = zip(*(
                self.__compute_matrices_single(I, Io, alpha, beta)
                for I in Is
            ))

            self.HERef = torch.stack(HEs).mean(dim=0)
            self.maxCRef = torch.stack(maxCs).mean(dim=0)
        elif self.norm_mode == 'concat':
            ODs, ODhats = zip(*(
                self.__convert_rgb2od(I, Io, beta)
                for I in Is
            ))
            OD = torch.cat(ODs, dim=0)
            ODhat = torch.cat(ODhats, dim=0)

            eigvecs =  torch.symeig(cov(ODhat.T), eigenvectors=True)[1][:, [1, 2]]

            HE =  self.__find_HE(ODhat, eigvecs, alpha)

            C = self.__find_concentration(OD, HE)
            maxCs = torch.stack([percentile(C[0, :], 99), percentile(C[1, :], 99)])
            self.HERef = HE
            self.maxCRef = maxCs
        elif self.norm_mode == 'avg-pre':
            ODs, ODhats = zip(*(
                self.__convert_rgb2od(I, Io, beta)
                for I in Is
            ))
            
            eigvecs = torch.stack([torch.symeig(cov(ODhat.T), eigenvectors=True)[1][:, [1, 2]] for ODhat in ODhats]).mean(dim=0)

            OD = torch.cat(ODs, dim=0)
            ODhat = torch.cat(ODhats, dim=0)
            
            HE =  self.__find_HE(ODhat, eigvecs, alpha)

            C = self.__find_concentration(OD, HE)
            maxCs = torch.stack([percentile(C[0, :], 99), percentile(C[1, :], 99)])
            self.HERef = HE
            self.maxCRef = maxCs
        elif self.norm_mode == 'fixed-single' or self.norm_mode == 'stochastic-single':
            # single img
            self.HERef, _, self.maxCRef = self.__compute_matrices_single(Is[0], Io, alpha, beta)
        else:
            raise "Unknown norm mode"

    def normalize(self, I, Io=240, alpha=1, beta=0.15, stains=True):
        c, h, w = I.shape

        HE, C, maxC = self.__compute_matrices_single(I, Io, alpha, beta)
        C = (self.maxCRef / maxC).unsqueeze(-1) * C

        Inorm = Io * torch.exp(-torch.matmul(self.HERef, C))
        Inorm[Inorm > 255] = 255
        Inorm = Inorm.T.reshape(h, w, c).int()

        H, E = None, None

        if stains:
            H = torch.mul(Io, torch.exp(torch.matmul(-self.HERef[:, 0].unsqueeze(-1), C[0, :].unsqueeze(0))))
            H[H > 255] = 255
            H = H.T.reshape(h, w, c).int()

            E = torch.mul(Io, torch.exp(torch.matmul(-self.HERef[:, 1].unsqueeze(-1), C[1, :].unsqueeze(0))))
            E[E > 255] = 255
            E = E.T.reshape(h, w, c).int()

        return Inorm, H, E


import torch
from torchstain.base.augmentors.he_augmentor import HEAugmentor
from torchstain.torch.utils import cov, percentile

"""
Source code ported from: https://github.com/schaugf/HEnorm_python
Original implementation: https://github.com/mitkovetta/staining-normalization
"""
class TorchMacenkoAugmentor(HEAugmentor):
    def __init__(self, sigma1=0.2, sigma2=0.2):
        super().__init__()

        self.sigma1 = sigma1
        self.sigma2 = sigma2

        self.I = None

        self.HERef = torch.tensor([[0.5626, 0.2159],
                                   [0.7201, 0.8012],
                                   [0.4062, 0.5581]])
        self.maxCRef = torch.tensor([1.9705, 1.0308])

        # Avoid using deprecated torch.lstsq (since 1.9.0)
        self.updated_lstsq = hasattr(torch.linalg, 'lstsq')
        
    def __convert_rgb2od(self, I, Io, beta):
        I = I.permute(1, 2, 0)

        # calculate optical density
        OD = -torch.log((I.reshape((-1, I.shape[-1])).float() + 1)/Io)

        # remove transparent pixels
        ODhat = OD[~torch.any(OD < beta, dim=1)]

        return OD, ODhat

    def __find_HE(self, ODhat, eigvecs, alpha):
        # project on the plane spanned by the eigenvectors corresponding to the two
        # largest eigenvalues
        That = torch.matmul(ODhat, eigvecs)
        phi = torch.atan2(That[:, 1], That[:, 0])

        minPhi = percentile(phi, alpha)
        maxPhi = percentile(phi, 100 - alpha)

        vMin = torch.matmul(eigvecs, torch.stack((torch.cos(minPhi), torch.sin(minPhi)))).unsqueeze(1)
        vMax = torch.matmul(eigvecs, torch.stack((torch.cos(maxPhi), torch.sin(maxPhi)))).unsqueeze(1)

        # a heuristic to make the vector corresponding to hematoxylin first and the
        # one corresponding to eosin second
        HE = torch.where(vMin[0] > vMax[0], torch.cat((vMin, vMax), dim=1), torch.cat((vMax, vMin), dim=1))

        return HE

    def __find_concentration(self, OD, HE):
        # rows correspond to channels (RGB), columns to OD values
        Y = OD.T

        # determine concentrations of the individual stains
        if not self.updated_lstsq:
            return torch.lstsq(Y, HE)[0][:2]
    
        return torch.linalg.lstsq(HE, Y)[0]

    def __compute_matrices(self, I, Io, alpha, beta):
        OD, ODhat = self.__convert_rgb2od(I, Io=Io, beta=beta)

        # compute eigenvectors
        _, eigvecs = torch.linalg.eigh(cov(ODhat.T)) 
        eigvecs = eigvecs[:, [1, 2]]

        HE = self.__find_HE(ODhat, eigvecs, alpha)

        C = self.__find_concentration(OD, HE)
        maxC = torch.stack([percentile(C[0, :], 99), percentile(C[1, :], 99)])

        return HE, C, maxC

    def fit(self, I, Io=240, alpha=1, beta=0.15):
        HE, C, maxC = self.__compute_matrices(I, Io, alpha, beta)

        # keep these as we will use them for augmentation
        self.I = I
        self.HERef = HE
        self.CRef = C
        self.maxCRef = maxC
    
    @staticmethod
    def random_uniform(shape, low, high):
        return (low - high) * torch.rand(*shape) + high

    def augment(self, Io=240, alpha=1, beta=0.15):
        I = self.I
        c, h, w = I.shape

        HE, C, maxC = self.__compute_matrices(I, Io, alpha, beta)

        maxC = maxC / self.maxCRef
        C2 = C / torch.unsqueeze(maxC, axis=-1)

        # introduce noise to the concentrations (applied along axis=0)
        C2 = (C2 * self.random_uniform((2, 1), 1 - self.sigma1, 1 + self.sigma1)) + self.random_uniform((2, 1), -self.sigma2, self.sigma2)

        # recreate the image using reference mixing matrix
        Inorm = Io * torch.exp(-torch.matmul(self.HERef, C2))
        Inorm[Inorm > 255] = 255
        Inorm = Inorm.T.reshape(h, w, c).int()

        return Inorm
        


import torch

def csplit(I):
    return [I[i] for i in range(I.shape[0])]

def cmerge(I1, I2, I3):
    return torch.stack([I1, I2, I3], dim=0)

def lab_split(I):
    I = I.type(torch.float32)
    I1, I2, I3 = csplit(I)
    return I1 / 2.55, I2 - 128, I3 - 128

def lab_merge(I1, I2, I3):
    return cmerge(I1 * 2.55, I2 + 128, I3 + 128)



from torchstain.torch.utils.cov import cov
from torchstain.torch.utils.percentile import percentile
from torchstain.torch.utils.stats import *
from torchstain.torch.utils.split import *
from torchstain.torch.utils.rgb2lab import *
from torchstain.torch.utils.lab2rgb import *



def ReinhardNormalizer(backend='numpy', method=None):
    if backend == 'numpy':
        from torchstain.numpy.normalizers import NumpyReinhardNormalizer
        return NumpyReinhardNormalizer(method=method)
    elif backend == "torch":
        from torchstain.torch.normalizers import TorchReinhardNormalizer
        return TorchReinhardNormalizer(method=method)
    elif backend == "tensorflow":
        from torchstain.tf.normalizers import TensorFlowReinhardNormalizer
        return TensorFlowReinhardNormalizer(method=method)
    else:
        raise Exception(f'Unknown backend {backend}')



import tensorflow as tf
from torchstain.base.augmentors.he_augmentor import HEAugmentor
from torchstain.tf.utils import cov, percentile, solveLS
import numpy as np
import tensorflow.keras.backend as K


"""
Source code ported from: https://github.com/schaugf/HEnorm_python
Original implementation: https://github.com/mitkovetta/staining-normalization
"""
class TensorFlowMacenkoAugmentor(HEAugmentor):
    def __init__(self, sigma1=0.2, sigma2=0.2):
        super().__init__()

        self.sigma1 = sigma1
        self.sigma2 = sigma2

        self.I = None

        self.HERef = tf.constant([[0.5626, 0.2159],
                                  [0.7201, 0.8012],
                                  [0.4062, 0.5581]])
        self.maxCRef = tf.constant([1.9705, 1.0308])

    def __convert_rgb2od(self, I, Io, beta):
        I = tf.transpose(I, [1, 2, 0])

        # calculate optical density
        OD = -tf.math.log((tf.cast(tf.reshape(I, [tf.math.reduce_prod(I.shape[:-1]), I.shape[-1]]), tf.float32) + 1) / Io)

        # remove transparent pixels
        ODhat = OD[~tf.math.reduce_any(OD < beta, axis=1)]

        return OD, ODhat

    def __find_HE(self, ODhat, eigvecs, alpha):
        # project on the plane spanned by the eigenvectors corresponding to the two
        # largest eigenvalues
        That = tf.linalg.matmul(ODhat, eigvecs)
        phi = tf.math.atan2(That[:, 1], That[:, 0])

        minPhi = percentile(phi, alpha)
        maxPhi = percentile(phi, 100 - alpha)

        vMin = tf.matmul(eigvecs, tf.expand_dims(tf.stack((tf.math.cos(minPhi), tf.math.sin(minPhi))), axis=-1))
        vMax = tf.matmul(eigvecs, tf.expand_dims(tf.stack((tf.math.cos(maxPhi), tf.math.sin(maxPhi))), axis=-1))

        # a heuristic to make the vector corresponding to hematoxylin first and the
        # one corresponding to eosin second
        HE = tf.where(vMin[0] > vMax[0], tf.concat((vMin, vMax), axis=1), tf.concat((vMax, vMin), axis=1))

        return HE

    def __find_concentration(self, OD, HE):
        # rows correspond to channels (RGB), columns to OD values
        Y = tf.transpose(OD)

        # determine concentrations of the individual stains
        return solveLS(HE, Y)[:2]

    def __compute_matrices(self, I, Io, alpha, beta):
        OD, ODhat = self.__convert_rgb2od(I, Io=Io, beta=beta)

        # compute eigenvectors
        _, eigvecs = tf.linalg.eigh(cov(tf.transpose(ODhat)))
        eigvecs = eigvecs[:, 1:3]

        HE = self.__find_HE(ODhat, eigvecs, alpha)

        C = self.__find_concentration(OD, HE)
        maxC = tf.stack([percentile(C[0, :], 99), percentile(C[1, :], 99)])

        return HE, C, maxC

    def fit(self, I, Io=240, alpha=1, beta=0.15):
        HE, C, maxC = self.__compute_matrices(I, Io, alpha, beta)

        # keep these as we will use them for augmentation
        self.I = I
        self.HERef = HE
        self.CRef = C
        self.maxCRef = maxC

    def augment(self, Io=240, alpha=1, beta=0.15):
        I = self.I
        c, h, w = I.shape

        HE, C, maxC = self.__compute_matrices(I, Io, alpha, beta)

        maxC = maxC / self.maxCRef
        C2 = C / tf.expand_dims(maxC, axis=-1)

        # introduce noise to the concentrations (applied along axis=0)
        C2 = (C2 * tf.random.uniform((2, 1), 1 - self.sigma1, 1 + self.sigma1)) + tf.random.uniform((2, 1), -self.sigma2, self.sigma2)

        # recreate the image using reference mixing matrix
        Iaug = Io * tf.math.exp(-tf.linalg.matmul(self.HERef, C2))
        Iaug = tf.clip_by_value(Iaug, 0, 255)
        Iaug = tf.cast(tf.reshape(tf.transpose(Iaug), shape=(h, w, c)), tf.int32)

        return Iaug



import torch

# constant conversion matrices between color spaces: https://gist.github.com/bikz05/6fd21c812ef6ebac66e1
_rgb2xyz = torch.tensor([[0.412453, 0.357580, 0.180423],
                         [0.212671, 0.715160, 0.072169],
                         [0.019334, 0.119193, 0.950227]])
                        
_white = torch.tensor([0.95047, 1., 1.08883])

def rgb2lab(rgb):
    arr = rgb.type(torch.float32)

    # convert rgb -> xyz color domain
    mask = arr > 0.04045
    not_mask = torch.logical_not(mask)
    arr.masked_scatter_(mask, torch.pow((torch.masked_select(arr, mask) + 0.055) / 1.055, 2.4))
    arr.masked_scatter_(not_mask, torch.masked_select(arr, not_mask) / 12.92)

    xyz = torch.tensordot(torch.t(_rgb2xyz), arr, dims=([0], [0]))

    # scale by CIE XYZ tristimulus values of the reference white point
    arr = torch.mul(xyz, 1 / _white.type(xyz.dtype).unsqueeze(dim=-1).unsqueeze(dim=-1))

    # nonlinear distortion and linear transformation
    mask = arr > 0.008856
    not_mask = torch.logical_not(mask)
    arr.masked_scatter_(mask, torch.pow(torch.masked_select(arr, mask), 1 / 3))
    arr.masked_scatter_(not_mask, 7.787 * torch.masked_select(arr, not_mask) + 16 / 116)

    # get each channel as individual tensors
    x, y, z = arr[0], arr[1], arr[2]

    # vector scaling
    L = (116. * y) - 16.
    a = 500.0 * (x - y)
    b = 200.0 * (y - z)

    # OpenCV format
    L *= 2.55
    a += 128
    b += 128

    # finally, get LAB color domain
    return torch.stack([L, a, b], axis=0)



from typing import Union
import tensorflow as tf

def percentile(t: tf.Tensor, q: float) -> Union[int, float]:
    """
    Return the ``q``-th percentile of the flattened input tensor's data.

    CAUTION:
     * Values are not interpolated, which corresponds to
       ``numpy.percentile(..., interpolation="nearest")``.

    :param t: Input tensor.
    :param q: Percentile to compute, which must be between 0 and 100 inclusive.
    :return: Resulting value (scalar).
    """
    k = 1 + tf.math.round(.01 * tf.cast(q, tf.float32) * (tf.cast(tf.math.reduce_prod(tf.size(t)), tf.float32) - 1))
    return tf.sort(tf.reshape(t, [-1]))[tf.cast(k - 1, tf.int32)]



import numpy as np
from torchstain.numpy.utils.rgb2lab import _rgb2xyz

_xyz2rgb = np.linalg.inv(_rgb2xyz)

"""
Implementation is based on:
https://github.com/scikit-image/scikit-image/blob/00177e14097237ef20ed3141ed454bc81b308f82/skimage/color/colorconv.py#L704
"""
def lab2rgb(lab: np.ndarray) -> np.ndarray:
    """
    Convert an array of LAB values to RGB values.

    Args:
        lab (np.ndarray): An array of shape (..., 3) containing LAB values.

    Returns:
        np.ndarray: An array of shape (..., 3) containing RGB values.
    """
    # first rescale back from OpenCV format
    lab[..., 0] /= 2.55
    lab[..., 1:] -= 128

    # convert LAB -> XYZ color domain
    y = (lab[..., 0] + 16.) / 116.
    x = (lab[..., 1] / 500.) + y
    z = y - (lab[..., 2] / 200.)

    xyz = np.stack([x, y, z], axis=-1)

    mask = xyz > 0.2068966
    xyz[mask] = np.power(xyz[mask], 3.)
    xyz[~mask] = (xyz[~mask] - 16.0 / 116.) / 7.787

    # rescale to the reference white (illuminant)
    xyz *= np.array((0.95047, 1., 1.08883), dtype=xyz.dtype)

    # convert XYZ -> RGB color domain
    rgb = np.matmul(xyz, _xyz2rgb.T)

    mask = rgb > 0.0031308
    rgb[mask] = 1.055 * np.power(rgb[mask], 1 / 2.4) - 0.055
    rgb[~mask] *= 12.92

    return np.clip(rgb, 0, 1)



import pathlib
from setuptools import setup, find_packages, find_namespace_packages

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name='torchstain',
    version='1.4.0',
    description='Stain normalization tools for histological analysis and computational pathology',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/EIDOSlab/torchstain',
    author='Carlo Alberto Barbano, André Pedersen',
    author_email='eidoslab@di.unito.it',
    license='MIT',
    packages=find_packages(exclude=('tests')),
    zip_safe=False,
    install_requires=[
        'numpy'
    ],
    extras_require={
        "tf": ["tensorflow"],
        "torch": ["torch"],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    python_requires='>=3.6'
)



import tensorflow as tf

def cov(x):
    """
    https://en.wikipedia.org/wiki/Covariance_matrix
    """
    E_x = tf.math.reduce_mean(x, axis=1)
    x = x - E_x[:, None]
    return tf.linalg.matmul(x, tf.transpose(x)) / (x.shape[1] - 1)



import os
import cv2
import torchstain
import torchstain.tf
import tensorflow as tf
import numpy as np

def test_cov():
    x = np.random.randn(10, 10)
    cov_np = np.cov(x)
    cov_t = torchstain.tf.utils.cov(x)

    np.testing.assert_almost_equal(cov_np, cov_t.numpy())

def test_percentile():
    x = np.random.randn(10, 10)
    p = 20
    p_np = np.percentile(x, p, interpolation='nearest')
    p_t = torchstain.tf.utils.percentile(x, p)

    np.testing.assert_almost_equal(p_np, p_t)

def test_macenko_tf():
    size = 1024
    curr_file_path = os.path.dirname(os.path.realpath(__file__))
    target = cv2.resize(cv2.cvtColor(cv2.imread(os.path.join(curr_file_path, "../data/target.png")), cv2.COLOR_BGR2RGB), (size, size))
    to_transform = cv2.resize(cv2.cvtColor(cv2.imread(os.path.join(curr_file_path, "../data/source.png")), cv2.COLOR_BGR2RGB), (size, size))

    # setup preprocessing and preprocess image to be normalized
    T = lambda x: tf.convert_to_tensor(np.moveaxis(x, -1, 0).astype("float32"))  #  * 255
    t_to_transform = T(to_transform)

    # initialize normalizers for each backend and fit to target image
    normalizer = torchstain.normalizers.MacenkoNormalizer(backend='numpy')
    normalizer.fit(target)

    tf_normalizer = torchstain.normalizers.MacenkoNormalizer(backend='tensorflow')
    tf_normalizer.fit(T(target))

    # transform
    result_numpy, _, _ = normalizer.normalize(I=to_transform, stains=True)
    result_tf, _, _ = tf_normalizer.normalize(I=t_to_transform, stains=True)

    # convert to numpy and set dtype
    result_numpy = result_numpy.astype("float32") / 255.
    result_tf = result_tf.numpy().astype("float32") / 255.

    # assess whether the normalized images are identical across backends
    np.testing.assert_almost_equal(result_numpy.flatten(), result_tf.flatten(), decimal=2, verbose=True)

def test_reinhard_tf():
    size = 1024
    curr_file_path = os.path.dirname(os.path.realpath(__file__))
    target = cv2.resize(cv2.cvtColor(cv2.imread(os.path.join(curr_file_path, "../data/target.png")), cv2.COLOR_BGR2RGB), (size, size))
    to_transform = cv2.resize(cv2.cvtColor(cv2.imread(os.path.join(curr_file_path, "../data/source.png")), cv2.COLOR_BGR2RGB), (size, size))

    # setup preprocessing and preprocess image to be normalized
    T = lambda x: tf.convert_to_tensor(x, dtype=tf.float32)
    t_to_transform = T(to_transform)

    # initialize normalizers for each backend and fit to target image
    normalizer = torchstain.normalizers.ReinhardNormalizer(backend='numpy')
    normalizer.fit(target)

    tf_normalizer = torchstain.normalizers.ReinhardNormalizer(backend='tensorflow')
    tf_normalizer.fit(T(target))

    # transform
    result_numpy = normalizer.normalize(I=to_transform)
    result_tf = tf_normalizer.normalize(I=t_to_transform)

    # convert to numpy and set dtype
    result_numpy = result_numpy.astype("float32") / 255.
    result_tf = result_tf.numpy().astype("float32") / 255.

    # assess whether the normalized images are identical across backends
    np.testing.assert_almost_equal(result_numpy.flatten(), result_tf.flatten(), decimal=2, verbose=True)



import numpy as np

# constant conversion matrices between color spaces: https://gist.github.com/bikz05/6fd21c812ef6ebac66e1
_rgb2xyz = np.array([[0.412453, 0.357580, 0.180423],
                     [0.212671, 0.715160, 0.072169],
                     [0.019334, 0.119193, 0.950227]])

"""
Implementation adapted from:
https://gist.github.com/bikz05/6fd21c812ef6ebac66e1
https://github.com/scikit-image/scikit-image/blob/00177e14097237ef20ed3141ed454bc81b308f82/skimage/color/colorconv.py#L704
"""
def rgb2lab(rgb):
    rgb = rgb.astype("float32")

    # convert rgb -> xyz color domain
    arr = rgb.copy()
    mask = arr > 0.04045
    arr[mask] = np.power((arr[mask] + 0.055) / 1.055, 2.4)
    arr[~mask] /= 12.92
    xyz = np.dot(arr, _rgb2xyz.T.astype(arr.dtype))

    # scale by CIE XYZ tristimulus values of the reference white point
    arr = xyz.copy()
    arr = arr / np.asarray((0.95047, 1., 1.08883), dtype=xyz.dtype)

    # Nonlinear distortion and linear transformation
    mask = arr > 0.008856
    arr[mask] = np.cbrt(arr[mask])
    arr[~mask] = 7.787 * arr[~mask] + 16. / 116.

    x, y, z = arr[..., 0], arr[..., 1], arr[..., 2]

    # Vector scaling
    L = (116. * y) - 16.
    a = 500.0 * (x - y)
    b = 200.0 * (y - z)

    # OpenCV format
    L *= 2.55
    a += 128
    b += 128

    # finally, get LAB color domain
    return np.concatenate([x[..., np.newaxis] for x in [L, a, b]], axis=-1)



def MultiMacenkoNormalizer(backend='torch', **kwargs):
    if backend == 'torch':
        from torchstain.torch.normalizers.multitarget import MultiMacenkoNormalizer
        return MultiMacenkoNormalizer(**kwargs)
    else:
        raise Exception(f'Unsupported backend {backend}')
